from __future__ import annotations

from typing import TYPE_CHECKING

from platformdirs import user_cache_path

if TYPE_CHECKING:
    from pathlib import Path


def _get_user_cache_path() -> Path:
    return user_cache_path(appname="pynaa", ensure_exists=True)



from __future__ import annotations

from datetime import datetime

from pydantic import AnyUrl, BaseModel, ConfigDict, HttpUrl
from torf import Torrent

from ._enums import NyaaCategory


class ParentModel(BaseModel):
    """
    Parent model that stores global configuration
    All models ahead will inherit from this.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)


class Submitter(ParentModel):
    """User who submitted the torrent."""

    name: str
    """Username of the submitter."""
    url: HttpUrl
    """Profile URL of the submitter."""
    is_trusted: bool
    """Indicates whether the user is trusted (green) or not."""
    is_banned: bool
    """Indicates whether the user is banned or not."""


class NyaaTorrentPage(ParentModel):
    """Nyaa's torrent page."""

    id: int
    """Nyaa ID of the torrent (`https://nyaa.si/view/{id}`)."""

    url: HttpUrl
    """URL to the Nyaa torrent page (`https://nyaa.si/view/123456`)."""

    title: str
    """Title of the torrent."""

    category: NyaaCategory
    """Torrent category."""

    date: datetime
    """Date and time at which the torrent was submitted."""

    submitter: Submitter
    """User who submitted the torrent."""

    information: str
    """Information about the torrent."""

    seeders: int
    """Number of seeders."""

    leechers: int
    """Number of leechers."""

    completed: int
    """Number of completed downloads."""

    is_trusted: bool
    """Indicates whether the upload is trusted (green) or not."""

    is_remake: bool
    """
    Indicates whether the upload is a remake (red) or not.
    
    Note
    ----
    An upload can be both trusted and a remake, in which case,
    the remake takes priority, that is, `is_remake` will be `True`
    and `is_trusted` will be `False`.
    This is a current limitation that I don't know how to work around.
    """

    description: str
    """Torrent description."""

    torrent_file: HttpUrl
    """URL pointing to the `.torrent` file (`https://nyaa.si/download/123456.torrent`)"""

    magnet: AnyUrl
    """
    Magnet link of the torrent. 
    
    Note
    ----
    The magnet link provided by Nyaa is different from the one 
    you'll get if you simply generated it from the `.torrent` file itself
    because Nyaa strips away all trackers except it's own 
    and the ones listed [here](https://github.com/nyaadevs/nyaa/blob/master/trackers.txt).
    """

    torrent: Torrent
    """
    A [`torf.Torrent`](https://torf.readthedocs.io/en/latest/#torf.Torrent) object 
    representing the data stored in the `.torrent` file.
    """


__all__ = [
    "Submitter",
    "NyaaTorrentPage",
]



from ._compat import StrEnum


class NyaaCategory(StrEnum):
    """Nyaa categories"""

    ANIME_MUSIC_VIDEO = "Anime - Anime Music Video"
    ANIME_ENGLISH_TRANSLATED = "Anime - English-translated"
    ANIME_NON_ENGLISH_TRANSLATED = "Anime - Non-English-translated"
    ANIME_RAW = "Anime - Raw"

    AUDIO_LOSSLESS = "Audio - Lossless"
    AUDIO_LOSSY = "Audio - Lossy"

    LITERATURE_ENGLISH_TRANSLATED = "Literature - English-translated"
    LITERATURE_NON_ENGLISH_TRANSLATED = "Literature - Non-English-translated"
    LITERATURE_RAW = "Literature - Raw"

    LIVE_ACTION_ENGLISH_TRANSLATED = "Live Action - English-translated"
    LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO = "Live Action - Idol/Promotional Video"
    LIVE_ACTION_NON_ENGLISH_TRANSLATED = "Live Action - Non-English-translated"
    LIVE_ACTION_RAW = "Live Action - Raw"

    PICTURES_GRAPHICS = "Pictures - Graphics"
    PICTURES_PHOTOS = "Pictures - Photos"

    SOFTWARE_APPLICATIONS = "Software - Applications"
    SOFTWARE_GAMES = "Software - Games"



"""
Aside from providing it's own types, this module also re-exports the following for convenience:
- [AnyUrl](https://docs.pydantic.dev/latest/api/networks/#pydantic.networks.AnyUrl)
- [HttpUrl](https://docs.pydantic.dev/latest/api/networks/#pydantic.networks.HttpUrl)
- [Torrent](https://torf.readthedocs.io/en/latest/#torf.Torrent)
- [File](https://torf.readthedocs.io/en/latest/#torf.File)
- [Filepath](https://torf.readthedocs.io/en/latest/#torf.Filepath)
"""

from __future__ import annotations

from typing import Any

from pydantic import AnyUrl, HttpUrl
from torf import File, Filepath, Torrent
from typing_extensions import TypeAlias

HTTPXClientKwargs: TypeAlias = Any
"""Simple TypeAlias to refer to `httpx.Client()` kwargs"""

HTTPXAsyncClientKwargs: TypeAlias = Any
"""Simple TypeAlias to refer to `httpx.AsyncClient()` kwargs"""

__all__ = [
    "AnyUrl",
    "HttpUrl",
    "HTTPXAsyncClientKwargs",
    "HTTPXClientKwargs",
    "Torrent",
    "File",
    "Filepath",
]



"""Compatibility module to for older python versions"""

import sys

if sys.version_info[0:2] < (3, 11):
    from strenum import StrEnum
else:
    from enum import StrEnum


if sys.version_info[0:2] < (3, 10):
    import importlib_metadata as metadata
else:
    from importlib import metadata

__all__ = ["metadata", "StrEnum"]



from ._clients import AsyncNyaa, Nyaa
from ._enums import NyaaCategory
from ._exceptions import (
    CloseError,
    ConnectError,
    ConnectTimeout,
    CookieConflict,
    DecodingError,
    HTMLParsingError,
    HTTPError,
    HTTPStatusError,
    InvalidURL,
    LocalProtocolError,
    NetworkError,
    PoolTimeout,
    ProtocolError,
    ProxyError,
    ReadError,
    ReadTimeout,
    RemoteProtocolError,
    RequestError,
    RequestNotRead,
    ResponseNotRead,
    StreamClosed,
    StreamConsumed,
    StreamError,
    TimeoutException,
    TooManyRedirects,
    UnsupportedProtocol,
    ValidationError,
    WriteError,
    WriteTimeout,
)
from ._models import NyaaTorrentPage
from ._types import AnyUrl, File, Filepath, HttpUrl, Torrent
from ._version import _get_version

__version__ = _get_version()
__version_tuple__ = tuple(__version__.split("."))

__all__ = [
    # Clients
    "AsyncNyaa",
    "Nyaa",
    # Enums
    "NyaaCategory",
    # Types
    "AnyUrl",
    "HttpUrl",
    "Torrent",
    "File",
    "Filepath",
    # Models
    "NyaaTorrentPage",
    # Exceptions
    "CloseError",
    "ConnectError",
    "ConnectTimeout",
    "CookieConflict",
    "DecodingError",
    "HTMLParsingError",
    "HTTPError",
    "HTTPStatusError",
    "InvalidURL",
    "LocalProtocolError",
    "NetworkError",
    "PoolTimeout",
    "ProtocolError",
    "ProxyError",
    "ReadError",
    "ReadTimeout",
    "RemoteProtocolError",
    "RequestError",
    "RequestNotRead",
    "ResponseNotRead",
    "StreamClosed",
    "StreamConsumed",
    "StreamError",
    "TimeoutException",
    "TooManyRedirects",
    "UnsupportedProtocol",
    "WriteError",
    "WriteTimeout",
    "ValidationError",
]



from __future__ import annotations

from io import BytesIO
from typing import Any

from bs4 import BeautifulSoup
from hishel import CacheClient, FileStorage
from pydantic import validate_call
from pydantic_core import Url
from torf import Torrent

from .._exceptions import HTMLParsingError
from .._models import NyaaTorrentPage
from .._types import HTTPXClientKwargs
from .._utils import _get_user_cache_path


class Nyaa:
    def __init__(self, base_url: str = "https://nyaa.si", cache: bool = True, **kwargs: HTTPXClientKwargs) -> None:
        """
        Nyaa client.

        Parameters
        ----------
        base_url : str, optional
            The base URL of Nyaa. Default is `https://nyaa.si`.
            This is only used when a Nyaa ID is passed.
            If a full URL is passed then this gets ignored and the base_url is parsed from the given URL instead.
        cache : bool, optional
            Whether to enable caching. Default is `True`.
            This will cache the page upon it's first request and then use the cached result
            for any subsequent requests for the same page.
            This helps in avoiding [HTTP 429 Error](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/429) but
            do note some fields like seeders, leechers, and completed are constantly changing and thus caching would
            mean you won't get the latest data on said fields.
        kwargs : HTTPXClientKwargs, optional
            Keyword arguments to pass to the underlying [httpx.Client()](https://www.python-httpx.org/api/#client)
            used to make the GET request.
        """
        self.base_url = base_url.strip("/")
        self.cache = cache
        self.kwargs = kwargs

    def _parse_nyaa(self, html: str) -> dict[str, Any]:
        """
        Parse the HTML page to extract relevant information

        Parameters
        ----------
        html : str
            HTML content of the Nyaa page.

        Raises
        ------
        HTMLParsingError
            Raised for any error encountered while parsing Nyaa's HTML content

        Returns
        -------
        dict[str, Any]
            A dictionary with the relevant information.
        """

        try:
            soup = BeautifulSoup(html, "lxml")

            if success := soup.find("div", class_="panel panel-success"):
                # Trusted uploads use `panel-success` for their green color
                body = success
                is_trusted = True
                is_remake = False
            elif remake := soup.find("div", class_="panel panel-danger"):
                # Remake uploads use `panel-danger` for their red color
                body = remake
                is_trusted = False
                is_remake = True
            else:
                # Default uploads use `panel-default`
                body = soup.find("div", class_="panel panel-default")  # type: ignore
                is_trusted = False
                is_remake = False

            title = soup.title.string.strip(":: Nyaa").strip()  # type: ignore

            rows = body.find("div", class_="panel-body").find_all("div", class_="row")  # type: ignore

            # ROW ONE
            row_one = rows[0].find_all("div", class_="col-md-5")
            category = row_one[0].get_text().strip()
            date = row_one[1]["data-timestamp"]

            # ROW TWO
            row_two = rows[1].find_all("div", class_="col-md-5")
            submitter = row_two[0].get_text().strip()
            if submitter.lower() != "anonymous":
                submitter_url = f"{self.base_url}{row_two[0].find('a').get('href', f'/user/{submitter}')}"
                submitter_status = row_two[0].find("a").get("title", None)

                if submitter_status is not None:
                    if submitter_status.lower() == "trusted":
                        submitter_trusted = True
                        submitter_banned = False
                    elif submitter_status.lower() == "banned user":
                        submitter_trusted = False
                        submitter_banned = True
                    elif submitter_status.lower() == "banned trusted":
                        submitter_trusted = True
                        submitter_banned = True
                    else:
                        submitter_trusted = False
                        submitter_banned = False
            else:
                submitter_url = self.base_url
                submitter_trusted = False
                submitter_banned = False

            seeders = row_two[1].get_text().strip()

            # ROW THREE
            row_three = rows[2].find_all("div", class_="col-md-5")
            information = row_three[0].get_text().strip()
            leechers = row_three[1].get_text().strip()

            # ROW FOUR
            row_four = rows[3].find_all("div", class_="col-md-5")
            completed = row_four[1].get_text().strip()

            # ROW FOOTER
            footer = body.find("div", class_="panel-footer clearfix").find_all("a")  # type: ignore
            torrent_file = f"{self.base_url}{footer[0]['href']}"
            magnet = footer[1]["href"]

            # DESCRIPTION
            description = soup.find(id="torrent-description").get_text().strip()  # type: ignore

            return dict(
                title=title,
                category=category,
                date=date,
                submitter=dict(
                    name=submitter,
                    url=submitter_url,
                    is_trusted=submitter_trusted,
                    is_banned=submitter_banned,
                ),
                information=information,
                seeders=seeders,
                leechers=leechers,
                completed=completed,
                is_trusted=is_trusted,
                is_remake=is_remake,
                description=description,
                torrent_file=torrent_file,
                magnet=magnet,
            )
        except Exception as error:
            raise HTMLParsingError("Encountered an error while parsing Nyaa's HTML content!") from error

    @validate_call
    def get(self, page: int | str) -> NyaaTorrentPage:
        """
        Retrieve information from a Nyaa torrent page.

        Parameters
        ----------
        page : int or str
            The torrent page.
            This can either be a URL like `https://nyaa.si/view/123456`
            or just the ID like `123456`

        Raises
        ------
        ValidationError
            Invalid input
        HTTPStatusError
            Nyaa returned a non 2xx response.
        HTMLParsingError
            Raised for any error encountered while parsing Nyaa's HTML content

        Returns
        -------
        NyaaTorrentPage
            A NyaaTorrentPage object representing the retrieved data.
        """

        if isinstance(page, int):
            url = f"{self.base_url}/view/{page}"
            id = page
        else:
            url = page
            id = page.split("/")[-1]  # type: ignore
            host = Url(page).host
            self.base_url = f"https://{host}" if host is not None else "https://nyaa.si"

        with CacheClient(storage=FileStorage(base_path=_get_user_cache_path()), **self.kwargs) as client:
            extensions = {"force_cache": self.cache, "cache_disabled": not self.cache}
            nyaa = client.get(url, extensions=extensions).raise_for_status()

            info = self._parse_nyaa(nyaa.text)

            # Get the torrent file and convert it to a torf.Torrent object
            torrent = Torrent.read_stream(
                BytesIO(client.get(info["torrent_file"], extensions=extensions).raise_for_status().content)
            )

            return NyaaTorrentPage(
                id=id,  # type: ignore
                url=url,  # type: ignore
                torrent=torrent,
                **info,
            )
